#include <stdio.h>

#include <windows.h>
#include <tlhelp32.h>

#include "ttl_define.h"
#include "tt_command_line_analyzer.h"
#include "tt_process.h"
#include "tt_string.h"
#include "tt_util.h"


#define TOXYZZY_VERSION "1.0.1"


static void
output_usage( void )
{
  fprintf( stderr, "Usage: toxyzzy.exe [-a[-]] [-b buffer_name]\n" );
}

static void
output_version( void )
{
  fprintf( stderr, "toxyzzy version %s (%s)\n", TOXYZZY_VERSION, TTL_DATE_STRING.c_str() );
}


// -- server -------------------------------------------------------------
struct XyzzyServerParameter {
public:
  int    size_;
  DWORD  process_id_;
  HANDLE event_;
  HWND   window_;
  int    kill_ok_;
  char   data[1];
};


class XyzzyServer {
public:
  static const char Name[];

public:
  explicit XyzzyServer( void ) : map_( 0 ), base_( 0 ) {}
  ~XyzzyServer();

  bool Allocate( unsigned int size );

  XyzzyServerParameter& GetParameter( void ) const { return *reinterpret_cast<XyzzyServerParameter*>( base_ ); }
  char* GetData( void ) const { return this->GetParameter().data; }
  HANDLE GetHandle( void ) const { return map_; }

private:
  HANDLE map_;
  void*  base_;
};

const char
XyzzyServer::Name[] = "xyzzy-server v2";

XyzzyServer::~XyzzyServer()
{
  if ( base_ ) {
    ::UnmapViewOfFile( base_ );
  }
  if ( map_ ) {
    ::CloseHandle( map_ );
  }
}

bool
XyzzyServer::Allocate( unsigned int size )
{
  size += sizeof( XyzzyServerParameter );
  map_ = ::CreateFileMapping( INVALID_HANDLE_VALUE, 0, PAGE_READWRITE, 0, size, 0 );
  if ( NOT( map_ ) ) {
    return false;
  }
  base_ = ::MapViewOfFile( map_, FILE_MAP_WRITE, 0, 0, 0 );
  if ( NOT( base_ ) ) {
    return false;
  }
  this->GetParameter().size_       = size;
  this->GetParameter().process_id_ = 0;
  this->GetParameter().event_      = 0;
  this->GetParameter().window_     = 0;
  this->GetParameter().kill_ok_    = 0;
  return true;
}

// -- lookup -------------------------------------------------------------
class XyzzyServerLookup {
private:
  static BOOL CALLBACK EnumerateProcess( HWND window, LPARAM param );

public:
  explicit XyzzyServerLookup( void ) : window_( 0 ), event_( 0 ) {}

  bool Find( void );

  HWND GetWindow( void ) const { return window_; }
  HANDLE GetEvent( void ) const { return event_; }
  HANDLE GetProcess( void ) const { return process_; }

private:
  HWND   window_;
  HANDLE event_;
  HANDLE process_;
};

BOOL CALLBACK
XyzzyServerLookup::EnumerateProcess( HWND window, LPARAM param )
{
  XyzzyServerLookup& lookup = *reinterpret_cast<XyzzyServerLookup*>( param );

  HANDLE handle = ::GetProp( window, XyzzyServer::Name );
  if ( NOT( handle ) ) {
    return TRUE;
  }
  DWORD process_id = 0;
  ::GetWindowThreadProcessId( window, &process_id );
  if ( NOT( process_id ) ) {
    return TRUE;
  }

  HANDLE process_handle = ::OpenProcess( PROCESS_ALL_ACCESS, 0, process_id );
  if ( NOT( process_handle ) ) {
    return TRUE;
  }

  HANDLE duplicated_handle;
  int ret = ::DuplicateHandle( process_handle, handle,
                               ::GetCurrentProcess(), &duplicated_handle,
                               0, 0, DUPLICATE_SAME_ACCESS );
  if ( NOT( ret ) ) {
    return TRUE;
  }
  lookup.window_ = window;
  lookup.event_  = duplicated_handle;
  lookup.process_ = process_handle;
  return FALSE;
}

bool
XyzzyServerLookup::Find( void )
{
  window_ = 0;
  event_ = 0;
  process_ = 0;
  ::EnumWindows( XyzzyServerLookup::EnumerateProcess, LPARAM( this ) );
  return window_;
}


// -- main -----
int
main( int argc, char* argv[] )
{
  int ret;
  std::string lisp_arguments;
  {
    TtCommandLineOptionAnalyzer      analyzer;

    TtCommandLineOptionBoolean       version_option( "version" );
    version_option.AddOptionString( "v" );
    analyzer.Regist( version_option );

    TtCommandLineOptionSingleOperand buffer_name_option( "buffer_name" );
    buffer_name_option.AddOptionString( "b" );
    analyzer.Regist( buffer_name_option );

    TtCommandLineOptionBoolean always_create_new_buffer_option( "always_create_new_buffer" );
    always_create_new_buffer_option.AddOptionString( "a" );
    always_create_new_buffer_option.SetDeniable( true );
    analyzer.Regist( always_create_new_buffer_option );

    std::vector<std::string> rest;
    try {
      rest = analyzer.AnalyzeForConsole( argc - 1, argv + 1 );
    }
    catch ( TtCommandLineOptionException& e ) {
      fprintf( stderr, "R}hCIvVsłBinfo : %s\n", e.Dump().c_str() );
      output_usage();
      return 1;
    }

    if ( version_option.GetResult() ) {
      output_version();
      return 0;
    }

    if ( rest.size() != 0 ) {
      fprintf( stderr, "R}hCsłB\n" );
      output_usage();
      return 1;
    }

    if ( always_create_new_buffer_option.IsSpecified() ) {
      lisp_arguments.append( " :always-create-new-buffer " );
      lisp_arguments.append( always_create_new_buffer_option.GetResult() ? "t" : "nil");
    }
    if ( buffer_name_option.IsSpecified() ) {
      lisp_arguments.append( " :buffer-name \"" );
      lisp_arguments.append( buffer_name_option.GetSingleOperand() );
      lisp_arguments.append( "\"" );
    }
  }

  if ( ::GetFileType( ::GetStdHandle( STD_INPUT_HANDLE ) ) == FILE_TYPE_CHAR ) {
    fprintf( stderr, "W͂R\[͂łB\n" );
    return 1;
  }

  XyzzyServerLookup lookup;
  ret = lookup.Find();
  if ( NOT( ret ) ) {
    fprintf( stderr, "xyzzy ܂łB\n" );
    return 1;
  }

  HANDLE duplicated_handle;
  ret = ::DuplicateHandle(
    ::GetCurrentProcess(), ::GetStdHandle( STD_INPUT_HANDLE ),
    lookup.GetProcess(), &duplicated_handle, 0, FALSE, DUPLICATE_SAME_ACCESS );
  ::CloseHandle( lookup.GetProcess() );
  if ( NOT( ret ) ) {
    fprintf( stderr, "DuplicateHandle s܂B\n" );
    fprintf( stderr, "system message : %s\n", TtUtil::GetWindowsSystemErrorMessage( ::GetLastError() ).c_str() );
    return 1;
  }

  std::string expression;
  expression.append( "(toxyzzy-start " );
  expression.append( TtUtil::ToStringFrom( reinterpret_cast<int>( duplicated_handle ) ) );
  expression.append( 1, ' ' );
  expression.append( lisp_arguments );
  expression.append( ")\"" );

  XyzzyServer server;
  ret = server.Allocate( expression.size() + 1 );
  if ( NOT( ret ) ) {
    fprintf( stderr, "܂B\n" );
    return 1;
  }
  strcpy( server.GetData(), expression.c_str() );

  ret = ::WaitForSingleObject( lookup.GetEvent(), 60 * 1000 ) == WAIT_OBJECT_0;
  ::CloseHandle( lookup.GetEvent() );
  if ( NOT( ret ) ) {
    fprintf( stderr, "xyzzyɐڑł܂B\n" );
    return 1;
  }

  ret = ::SendMessage( lookup.GetWindow(),
                       ::RegisterWindowMessage( XyzzyServer::Name ),
                       ::GetCurrentProcessId(),
                       LPARAM( server.GetHandle() ) );

  if ( NOT( ret ) ) {
    fprintf( stderr, "xyzzyƂ̒ʐMŃG[܂B\n" );
    return 1;
  }
  
  return 0;
}
